# -*- coding: utf-8 -*-
#
"""
Axiomatic triples to be (possibly) added to the final graph.

**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""

__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

from rdflib.namespace import OWL, RDF, RDFS, XSD

# Simple RDF axiomatic triples (typing of subject, predicate, first, rest, etc)
_Simple_RDF_axiomatic_triples = [
    (RDF.type, RDF.type, RDF.Property),
    (RDF.subject, RDF.type, RDF.Property),
    (RDF.predicate, RDF.type, RDF.Property),
    (RDF.object, RDF.type, RDF.Property),
    (RDF.first, RDF.type, RDF.Property),
    (RDF.rest, RDF.type, RDF.Property),
    (RDF.value, RDF.type, RDF.Property),
    (RDF.nil, RDF.type, RDF.List),
]

# RDFS axiomatic triples (domain and range, as well as class setting for a number of RDFS symbols)
_RDFS_axiomatic_triples = [
    (RDF.type, RDFS.domain, RDFS.Resource),
    (RDFS.domain, RDFS.domain, RDF.Property),
    (RDFS.range, RDFS.domain, RDF.Property),
    (RDFS.subPropertyOf, RDFS.domain, RDF.Property),
    (RDFS.subClassOf, RDFS.domain, RDFS.Class),
    (RDF.subject, RDFS.domain, RDF.Statement),
    (RDF.predicate, RDFS.domain, RDF.Statement),
    (RDF.object, RDFS.domain, RDF.Statement),
    (RDFS.member, RDFS.domain, RDFS.Resource),
    (RDF.first, RDFS.domain, RDF.List),
    (RDF.rest, RDFS.domain, RDF.List),
    (RDFS.seeAlso, RDFS.domain, RDFS.Resource),
    (RDFS.isDefinedBy, RDFS.domain, RDFS.Resource),
    (RDFS.comment, RDFS.domain, RDFS.Resource),
    (RDFS.label, RDFS.domain, RDFS.Resource),
    (RDF.value, RDFS.domain, RDFS.Resource),
    (RDF.Property, RDF.type, RDFS.Class),
    (RDF.type, RDFS.range, RDFS.Class),
    (RDFS.domain, RDFS.range, RDFS.Class),
    (RDFS.range, RDFS.range, RDFS.Class),
    (RDFS.subPropertyOf, RDFS.range, RDF.Property),
    (RDFS.subClassOf, RDFS.range, RDFS.Class),
    (RDF.subject, RDFS.range, RDFS.Resource),
    (RDF.predicate, RDFS.range, RDFS.Resource),
    (RDF.object, RDFS.range, RDFS.Resource),
    (RDFS.member, RDFS.range, RDFS.Resource),
    (RDF.first, RDFS.range, RDFS.Resource),
    (RDF.rest, RDFS.range, RDF.List),
    (RDFS.seeAlso, RDFS.range, RDFS.Resource),
    (RDFS.isDefinedBy, RDFS.range, RDFS.Resource),
    (RDFS.comment, RDFS.range, RDFS.Literal),
    (RDFS.label, RDFS.range, RDFS.Literal),
    (RDF.value, RDFS.range, RDFS.Resource),
    (RDF.Alt, RDFS.subClassOf, RDFS.Container),
    (RDF.Bag, RDFS.subClassOf, RDFS.Container),
    (RDF.Seq, RDFS.subClassOf, RDFS.Container),
    (RDFS.ContainerMembershipProperty, RDFS.subClassOf, RDF.Property),
    (RDFS.isDefinedBy, RDFS.subPropertyOf, RDFS.seeAlso),
    (RDF.XMLLiteral, RDF.type, RDFS.Datatype),
    (RDF.XMLLiteral, RDFS.subClassOf, RDFS.Literal),
    (RDFS.Datatype, RDFS.subClassOf, RDFS.Class),
    # rdfs valid triples; these would be inferred by the RDFS expansion, but it may make things
    # a bit faster to add these upfront
    (RDFS.Resource, RDF.type, RDFS.Class),
    (RDFS.Class, RDF.type, RDFS.Class),
    (RDFS.Literal, RDF.type, RDFS.Class),
    (RDF.XMLLiteral, RDF.type, RDFS.Class),
    (RDFS.Datatype, RDF.type, RDFS.Class),
    (RDF.Seq, RDF.type, RDFS.Class),
    (RDF.Bag, RDF.type, RDFS.Class),
    (RDF.Alt, RDF.type, RDFS.Class),
    (RDFS.Container, RDF.type, RDFS.Class),
    (RDF.List, RDF.type, RDFS.Class),
    (RDFS.ContainerMembershipProperty, RDF.type, RDFS.Class),
    (RDF.Property, RDF.type, RDFS.Class),
    (RDF.Statement, RDF.type, RDFS.Class),
    (RDFS.domain, RDF.type, RDF.Property),
    (RDFS.range, RDF.type, RDF.Property),
    (RDFS.subPropertyOf, RDF.type, RDF.Property),
    (RDFS.subClassOf, RDF.type, RDF.Property),
    (RDFS.member, RDF.type, RDF.Property),
    (RDFS.seeAlso, RDF.type, RDF.Property),
    (RDFS.isDefinedBy, RDF.type, RDF.Property),
    (RDFS.comment, RDF.type, RDF.Property),
    (RDFS.label, RDF.type, RDF.Property),
]

# RDFS Axiomatic Triples all together
RDFS_Axiomatic_Triples = _Simple_RDF_axiomatic_triples + _RDFS_axiomatic_triples

# RDFS D-entailement triples, ie, possible subclassing of various datatypes
RDFS_D_Axiomatic_Triples_subclasses = [
    # See http://www.w3.org/TR/2004/REC-xmlschema-2-20041028/#built-in-datatypes
    (XSD.decimal, RDFS.subClassOf, RDFS.Literal),
    (XSD.integer, RDFS.subClassOf, XSD.decimal),
    (XSD.long, RDFS.subClassOf, XSD.integer),
    (XSD.int, RDFS.subClassOf, XSD.long),
    (XSD.short, RDFS.subClassOf, XSD.int),
    (XSD.byte, RDFS.subClassOf, XSD.short),
    (XSD.nonNegativeInteger, RDFS.subClassOf, XSD.integer),
    (XSD.positiveInteger, RDFS.subClassOf, XSD.nonNegativeInteger),
    (XSD.unsignedLong, RDFS.subClassOf, XSD.nonNegativeInteger),
    (XSD.unsignedInt, RDFS.subClassOf, XSD.unsignedLong),
    (XSD.unsignedShort, RDFS.subClassOf, XSD.unsignedInt),
    (XSD.unsignedByte, RDFS.subClassOf, XSD.unsignedShort),
    (XSD.nonPositiveInteger, RDFS.subClassOf, XSD.integer),
    (XSD.negativeInteger, RDFS.subClassOf, XSD.nonPositiveInteger),
    (XSD.normalizedString, RDFS.subClassOf, XSD.string),
    (XSD.token, RDFS.subClassOf, XSD.normalizedString),
    (XSD.language, RDFS.subClassOf, XSD.token),
    (XSD.Name, RDFS.subClassOf, XSD.token),
    (XSD.NMTOKEN, RDFS.subClassOf, XSD.token),
    (XSD.NCName, RDFS.subClassOf, XSD.Name),
    (XSD.dateTimeStamp, RDFS.subClassOf, XSD.dateTime),
]

#
RDFS_D_Axiomatic_Triples_types = [
    (XSD.integer, RDF.type, RDFS.Datatype),
    (XSD.decimal, RDF.type, RDFS.Datatype),
    (XSD.nonPositiveInteger, RDF.type, RDFS.Datatype),
    (XSD.nonPositiveInteger, RDF.type, RDFS.Datatype),
    (XSD.positiveInteger, RDF.type, RDFS.Datatype),
    (XSD.positiveInteger, RDF.type, RDFS.Datatype),
    (XSD.long, RDF.type, RDFS.Datatype),
    (XSD.int, RDF.type, RDFS.Datatype),
    (XSD.short, RDF.type, RDFS.Datatype),
    (XSD.byte, RDF.type, RDFS.Datatype),
    (XSD.unsignedLong, RDF.type, RDFS.Datatype),
    (XSD.unsignedInt, RDF.type, RDFS.Datatype),
    (XSD.unsignedShort, RDF.type, RDFS.Datatype),
    (XSD.unsignedByte, RDF.type, RDFS.Datatype),
    (XSD.float, RDF.type, RDFS.Datatype),
    (XSD.double, RDF.type, RDFS.Datatype),
    (XSD.string, RDF.type, RDFS.Datatype),
    (XSD.normalizedString, RDF.type, RDFS.Datatype),
    (XSD.token, RDF.type, RDFS.Datatype),
    (XSD.language, RDF.type, RDFS.Datatype),
    (XSD.Name, RDF.type, RDFS.Datatype),
    (XSD.NCName, RDF.type, RDFS.Datatype),
    (XSD.NMTOKEN, RDF.type, RDFS.Datatype),
    (XSD.boolean, RDF.type, RDFS.Datatype),
    (XSD.hexBinary, RDF.type, RDFS.Datatype),
    (XSD.base64Binary, RDF.type, RDFS.Datatype),
    (XSD.anyURI, RDF.type, RDFS.Datatype),
    (XSD.dateTimeStamp, RDF.type, RDFS.Datatype),
    (XSD.dateTime, RDF.type, RDFS.Datatype),
    (RDFS.Literal, RDF.type, RDFS.Datatype),
    (RDF.XMLLiteral, RDF.type, RDFS.Datatype),
]

RDFS_D_Axiomatic_Triples = (
    RDFS_D_Axiomatic_Triples_types + RDFS_D_Axiomatic_Triples_subclasses
)

# OWL Class axiomatic triples: definition of special classes
_OWL_axiomatic_triples_Classes = [
    (OWL.AllDifferent, RDF.type, RDFS.Class),
    (OWL.AllDifferent, RDFS.subClassOf, RDFS.Resource),
    (OWL.AllDisjointClasses, RDF.type, RDFS.Class),
    (OWL.AllDisjointClasses, RDFS.subClassOf, RDFS.Resource),
    (OWL.AllDisjointProperties, RDF.type, RDFS.Class),
    (OWL.AllDisjointProperties, RDFS.subClassOf, RDFS.Resource),
    (OWL.Annotation, RDF.type, RDFS.Class),
    (OWL.Annotation, RDFS.subClassOf, RDFS.Resource),
    (OWL.AnnotationProperty, RDF.type, RDFS.Class),
    (OWL.AnnotationProperty, RDFS.subClassOf, RDF.Property),
    (OWL.AsymmetricProperty, RDF.type, RDFS.Class),
    (OWL.AsymmetricProperty, RDFS.subClassOf, RDF.Property),
    (OWL.Class, RDF.type, RDFS.Class),
    (OWL.Class, OWL.equivalentClass, RDFS.Class),
    #    (OWL.DataRange, RDF.type, RDFS.Class),
    #    (OWL.DataRange, OWL.equivalentClass, RDFS.Datatype),
    (RDFS.Datatype, RDF.type, RDFS.Class),
    (OWL.DatatypeProperty, RDF.type, RDFS.Class),
    (OWL.DatatypeProperty, RDFS.subClassOf, RDF.Property),
    (OWL.DeprecatedClass, RDF.type, RDFS.Class),
    (OWL.DeprecatedClass, RDFS.subClassOf, RDFS.Class),
    (OWL.DeprecatedProperty, RDF.type, RDFS.Class),
    (OWL.DeprecatedProperty, RDFS.subClassOf, RDF.Property),
    (OWL.FunctionalProperty, RDF.type, RDFS.Class),
    (OWL.FunctionalProperty, RDFS.subClassOf, RDF.Property),
    (OWL.InverseFunctionalProperty, RDF.type, RDFS.Class),
    (OWL.InverseFunctionalProperty, RDFS.subClassOf, RDF.Property),
    (OWL.IrreflexiveProperty, RDF.type, RDFS.Class),
    (OWL.IrreflexiveProperty, RDFS.subClassOf, RDF.Property),
    (RDFS.Literal, RDF.type, RDFS.Datatype),
    #    (OWL.NamedIndividual, RDF.type, RDFS.Class),
    #    (OWL.NamedIndividual, OWL.equivalentClass, RDFS.Resource),
    (OWL.NegativePropertyAssertion, RDF.type, RDFS.Class),
    (OWL.NegativePropertyAssertion, RDFS.subClassOf, RDFS.Resource),
    (OWL.Nothing, RDF.type, RDFS.Class),
    (OWL.Nothing, RDFS.subClassOf, OWL.Thing),
    (OWL.ObjectProperty, RDF.type, RDFS.Class),
    (OWL.ObjectProperty, OWL.equivalentClass, RDF.Property),
    (OWL.Ontology, RDF.type, RDFS.Class),
    (OWL.Ontology, RDFS.subClassOf, RDFS.Resource),
    (OWL.OntologyProperty, RDF.type, RDFS.Class),
    (OWL.OntologyProperty, RDFS.subClassOf, RDF.Property),
    (RDF.Property, RDF.type, RDFS.Class),
    (OWL.ReflexiveProperty, RDF.type, RDFS.Class),
    (OWL.ReflexiveProperty, RDFS.subClassOf, RDF.Property),
    (OWL.Restriction, RDF.type, RDFS.Class),
    (OWL.Restriction, RDFS.subClassOf, RDFS.Class),
    (OWL.SymmetricProperty, RDF.type, RDFS.Class),
    (OWL.SymmetricProperty, RDFS.subClassOf, RDF.Property),
    (OWL.Thing, RDF.type, RDFS.Class),
    (OWL.Thing, RDFS.subClassOf, RDFS.Resource),
    (OWL.TransitiveProperty, RDF.type, RDFS.Class),
    (OWL.TransitiveProperty, RDFS.subClassOf, RDF.Property),
    # OWL valid triples; some of these would be inferred by the OWL RL expansion, but it may make things
    # a bit faster to add these upfront
    (OWL.AllDisjointProperties, RDF.type, OWL.Class),
    (OWL.AllDisjointClasses, RDF.type, OWL.Class),
    (OWL.AllDisjointProperties, RDF.type, OWL.Class),
    (OWL.Annotation, RDF.type, OWL.Class),
    (OWL.AsymmetricProperty, RDF.type, OWL.Class),
    (OWL.Axiom, RDF.type, OWL.Class),
    (OWL.DataRange, RDF.type, OWL.Class),
    (RDFS.Datatype, RDF.type, OWL.Class),
    (OWL.DatatypeProperty, RDF.type, OWL.Class),
    (OWL.DeprecatedClass, RDF.type, OWL.Class),
    (OWL.DeprecatedClass, RDFS.subClassOf, OWL.Class),
    (OWL.DeprecatedProperty, RDF.type, OWL.Class),
    (OWL.FunctionalProperty, RDF.type, OWL.Class),
    (OWL.InverseFunctionalProperty, RDF.type, OWL.Class),
    (OWL.IrreflexiveProperty, RDF.type, OWL.Class),
    (OWL.NamedIndividual, RDF.type, OWL.Class),
    (OWL.NegativePropertyAssertion, RDF.type, OWL.Class),
    (OWL.Nothing, RDF.type, OWL.Class),
    (OWL.ObjectProperty, RDF.type, OWL.Class),
    (OWL.Ontology, RDF.type, OWL.Class),
    (OWL.OntologyProperty, RDF.type, OWL.Class),
    (RDF.Property, RDF.type, OWL.Class),
    (OWL.ReflexiveProperty, RDF.type, OWL.Class),
    (OWL.Restriction, RDF.type, OWL.Class),
    (OWL.Restriction, RDFS.subClassOf, OWL.Class),
    #    (OWL.SelfRestriction, RDF.type, OWL.Class),
    (OWL.SymmetricProperty, RDF.type, OWL.Class),
    (OWL.Thing, RDF.type, OWL.Class),
    (OWL.TransitiveProperty, RDF.type, OWL.Class),
]

# OWL Property axiomatic triples: definition of domains and ranges
_OWL_axiomatic_triples_Properties = [
    (OWL.allValuesFrom, RDF.type, RDF.Property),
    (OWL.allValuesFrom, RDFS.domain, OWL.Restriction),
    (OWL.allValuesFrom, RDFS.range, RDFS.Class),
    (OWL.assertionProperty, RDF.type, RDF.Property),
    (OWL.assertionProperty, RDFS.domain, OWL.NegativePropertyAssertion),
    (OWL.assertionProperty, RDFS.range, RDF.Property),
    (OWL.backwardCompatibleWith, RDF.type, OWL.OntologyProperty),
    (OWL.backwardCompatibleWith, RDF.type, OWL.AnnotationProperty),
    (OWL.backwardCompatibleWith, RDFS.domain, OWL.Ontology),
    (OWL.backwardCompatibleWith, RDFS.range, OWL.Ontology),
    #    (OWL.bottomDataProperty, RDF.type, RDFS.DatatypeProperty),
    #    (OWL.bottomObjectProperty, RDF.type, OWL.ObjectProperty),
    #    (OWL.cardinality, RDF.type, RDF.Property),
    #    (OWL.cardinality, RDFS.domain, OWL.Restriction),
    #    (OWL.cardinality, RDFS.range, XSD.nonNegativeInteger),
    (RDFS.comment, RDF.type, OWL.AnnotationProperty),
    (RDFS.comment, RDFS.domain, RDFS.Resource),
    (RDFS.comment, RDFS.range, RDFS.Literal),
    (OWL.complementOf, RDF.type, RDF.Property),
    (OWL.complementOf, RDFS.domain, RDFS.Class),
    (OWL.complementOf, RDFS.range, RDFS.Class),
    #    (OWL.datatypeComplementOf, RDF.type, RDF.Property),
    #    (OWL.datatypeComplementOf, RDFS.domain, RDFS.Datatype),
    #    (OWL.datatypeComplementOf, RDFS.range, RDFS.Datatype),
    (OWL.deprecated, RDF.type, OWL.AnnotationProperty),
    (OWL.deprecated, RDFS.domain, RDFS.Resource),
    (OWL.deprecated, RDFS.range, RDFS.Resource),
    (OWL.differentFrom, RDF.type, RDF.Property),
    (OWL.differentFrom, RDFS.domain, RDFS.Resource),
    (OWL.differentFrom, RDFS.range, RDFS.Resource),
    #    (OWL.disjointUnionOf, RDF.type, RDF.Property),
    #    (OWL.disjointUnionOf, RDFS.domain, RDFS.Class),
    #    (OWL.disjointUnionOf, RDFS.range, RDF.List),
    (OWL.disjointWith, RDF.type, RDF.Property),
    (OWL.disjointWith, RDFS.domain, RDFS.Class),
    (OWL.disjointWith, RDFS.range, RDFS.Class),
    (OWL.distinctMembers, RDF.type, RDF.Property),
    (OWL.distinctMembers, RDFS.domain, OWL.AllDifferent),
    (OWL.distinctMembers, RDFS.range, RDF.List),
    (OWL.equivalentClass, RDF.type, RDF.Property),
    (OWL.equivalentClass, RDFS.domain, RDFS.Class),
    (OWL.equivalentClass, RDFS.range, RDFS.Class),
    (OWL.equivalentProperty, RDF.type, RDF.Property),
    (OWL.equivalentProperty, RDFS.domain, RDF.Property),
    (OWL.equivalentProperty, RDFS.range, RDF.Property),
    (OWL.hasKey, RDF.type, RDF.Property),
    (OWL.hasKey, RDFS.domain, RDFS.Class),
    (OWL.hasKey, RDFS.range, RDF.List),
    (OWL.hasValue, RDF.type, RDF.Property),
    (OWL.hasValue, RDFS.domain, OWL.Restriction),
    (OWL.hasValue, RDFS.range, RDFS.Resource),
    (OWL.imports, RDF.type, OWL.OntologyProperty),
    (OWL.imports, RDFS.domain, OWL.Ontology),
    (OWL.imports, RDFS.range, OWL.Ontology),
    (OWL.incompatibleWith, RDF.type, OWL.OntologyProperty),
    (OWL.incompatibleWith, RDF.type, OWL.AnnotationProperty),
    (OWL.incompatibleWith, RDFS.domain, OWL.Ontology),
    (OWL.incompatibleWith, RDFS.range, OWL.Ontology),
    (OWL.intersectionOf, RDF.type, RDF.Property),
    (OWL.intersectionOf, RDFS.domain, RDFS.Class),
    (OWL.intersectionOf, RDFS.range, RDF.List),
    (OWL.inverseOf, RDF.type, RDF.Property),
    (OWL.inverseOf, RDFS.domain, RDF.Property),
    (OWL.inverseOf, RDFS.range, RDF.Property),
    (RDFS.isDefinedBy, RDF.type, OWL.AnnotationProperty),
    (RDFS.isDefinedBy, RDFS.domain, RDFS.Resource),
    (RDFS.isDefinedBy, RDFS.range, RDFS.Resource),
    (RDFS.label, RDF.type, OWL.AnnotationProperty),
    (RDFS.label, RDFS.domain, RDFS.Resource),
    (RDFS.label, RDFS.range, RDFS.Literal),
    (OWL.maxCardinality, RDF.type, RDF.Property),
    (OWL.maxCardinality, RDFS.domain, OWL.Restriction),
    (OWL.maxCardinality, RDFS.range, XSD.nonNegativeInteger),
    (OWL.maxQualifiedCardinality, RDF.type, RDF.Property),
    (OWL.maxQualifiedCardinality, RDFS.domain, OWL.Restriction),
    (OWL.maxQualifiedCardinality, RDFS.range, XSD.nonNegativeInteger),
    (OWL.members, RDF.type, RDF.Property),
    (OWL.members, RDFS.domain, RDFS.Resource),
    (OWL.members, RDFS.range, RDF.List),
    #    (OWL.minCardinality, RDF.type, RDF.Property),
    #    (OWL.minCardinality, RDFS.domain, OWL.Restriction),
    #    (OWL.minCardinality, RDFS.range, XSD.nonNegativeInteger),
    #    (OWL.minQualifiedCardinality, RDF.type, RDF.Property),
    #    (OWL.minQualifiedCardinality, RDFS.domain, OWL.Restriction),
    #    (OWL.minQualifiedCardinality, RDFS.range, XSD.nonNegativeInteger),
    #    (OWL.annotatedTarget, RDF.type, RDF.Property),
    #    (OWL.annotatedTarget, RDFS.domain, RDFS.Resource),
    #    (OWL.annotatedTarget, RDFS.range, RDFS.Resource),
    (OWL.onClass, RDF.type, RDF.Property),
    (OWL.onClass, RDFS.domain, OWL.Restriction),
    (OWL.onClass, RDFS.range, RDFS.Class),
    #    (OWL.onDataRange, RDF.type, RDF.Property),
    #    (OWL.onDataRange, RDFS.domain, OWL.Restriction),
    #    (OWL.onDataRange, RDFS.range, RDFS.Datatype),
    (OWL.onDatatype, RDF.type, RDF.Property),
    (OWL.onDatatype, RDFS.domain, RDFS.Datatype),
    (OWL.onDatatype, RDFS.range, RDFS.Datatype),
    (OWL.oneOf, RDF.type, RDF.Property),
    (OWL.oneOf, RDFS.domain, RDFS.Class),
    (OWL.oneOf, RDFS.range, RDF.List),
    (OWL.onProperty, RDF.type, RDF.Property),
    (OWL.onProperty, RDFS.domain, OWL.Restriction),
    (OWL.onProperty, RDFS.range, RDF.Property),
    #    (OWL.onProperties, RDF.type, RDF.Property),
    #    (OWL.onProperties, RDFS.domain, OWL.Restriction),
    #    (OWL.onProperties, RDFS.range, RDF.List),
    #    (OWL.annotatedProperty, RDF.type, RDF.Property),
    #    (OWL.annotatedProperty, RDFS.domain, RDFS.Resource),
    #    (OWL.annotatedProperty, RDFS.range, RDF.Property),
    (OWL.priorVersion, RDF.type, OWL.OntologyProperty),
    (OWL.priorVersion, RDF.type, OWL.AnnotationProperty),
    (OWL.priorVersion, RDFS.domain, OWL.Ontology),
    (OWL.priorVersion, RDFS.range, OWL.Ontology),
    (OWL.propertyChainAxiom, RDF.type, RDF.Property),
    (OWL.propertyChainAxiom, RDFS.domain, RDF.Property),
    (OWL.propertyChainAxiom, RDFS.range, RDF.List),
    #    (OWL.propertyDisjointWith, RDF.type, RDF.Property),
    #    (OWL.propertyDisjointWith, RDFS.domain, RDF.Property),
    #    (OWL.propertyDisjointWith, RDFS.range, RDF.Property),
    #
    #    (OWL.qualifiedCardinality, RDF.type, RDF.Property),
    #    (OWL.qualifiedCardinality, RDFS.domain, OWL.Restriction),
    #    (OWL.qualifiedCardinality, RDFS.range, XSD.nonNegativeInteger),
    (OWL.sameAs, RDF.type, RDF.Property),
    (OWL.sameAs, RDFS.domain, RDFS.Resource),
    (OWL.sameAs, RDFS.range, RDFS.Resource),
    (RDFS.seeAlso, RDF.type, OWL.AnnotationProperty),
    (RDFS.seeAlso, RDFS.domain, RDFS.Resource),
    (RDFS.seeAlso, RDFS.range, RDFS.Resource),
    (OWL.someValuesFrom, RDF.type, RDF.Property),
    (OWL.someValuesFrom, RDFS.domain, OWL.Restriction),
    (OWL.someValuesFrom, RDFS.range, RDFS.Class),
    (OWL.sourceIndividual, RDF.type, RDF.Property),
    (OWL.sourceIndividual, RDFS.domain, OWL.NegativePropertyAssertion),
    (OWL.sourceIndividual, RDFS.range, RDFS.Resource),
    #
    #    (OWL.annotatedSource, RDF.type, RDF.Property),
    #    (OWL.annotatedSource, RDFS.domain, RDFS.Resource),
    #    (OWL.annotatedSource, RDFS.range, RDFS.Resource),
    #
    (OWL.targetIndividual, RDF.type, RDF.Property),
    (OWL.targetIndividual, RDFS.domain, OWL.NegativePropertyAssertion),
    (OWL.targetIndividual, RDFS.range, RDFS.Resource),
    (OWL.targetValue, RDF.type, RDF.Property),
    (OWL.targetValue, RDFS.domain, OWL.NegativePropertyAssertion),
    (OWL.targetValue, RDFS.range, RDFS.Literal),
    #    (OWL.topDataProperty, RDF.type, RDFS.DatatypeProperty),
    #    (OWL.topDataProperty, RDFS.domain, RDFS.Resource),
    #    (OWL.topDataProperty, RDFS.range, RDFS.Literal),
    #
    #    (OWL.topObjectProperty, RDF.type, OWL.ObjectProperty),
    #    (OWL.topObjectProperty, RDFS.domain, RDFS.Resource),
    #    (OWL.topObjectProperty, RDFS.range, RDFS.Resource),
    (OWL.unionOf, RDF.type, RDF.Property),
    (OWL.unionOf, RDFS.domain, RDFS.Class),
    (OWL.unionOf, RDFS.range, RDF.List),
    (OWL.versionInfo, RDF.type, OWL.AnnotationProperty),
    (OWL.versionInfo, RDFS.domain, RDFS.Resource),
    (OWL.versionInfo, RDFS.range, RDFS.Resource),
    (OWL.versionIRI, RDF.type, OWL.AnnotationProperty),
    (OWL.versionIRI, RDFS.domain, RDFS.Resource),
    (OWL.versionIRI, RDFS.range, RDFS.Resource),
    (OWL.withRestrictions, RDF.type, RDF.Property),
    (OWL.withRestrictions, RDFS.domain, RDFS.Datatype),
    (OWL.withRestrictions, RDFS.range, RDF.List),
    # some OWL valid triples; these would be inferred by the OWL RL expansion, but it may make things
    # a bit faster to add these upfront
    (OWL.allValuesFrom, RDFS.range, OWL.Class),
    (OWL.complementOf, RDFS.domain, OWL.Class),
    (OWL.complementOf, RDFS.range, OWL.Class),
    #    (OWL.datatypeComplementOf, domain, OWL.DataRange),
    #    (OWL.datatypeComplementOf, range, OWL.DataRange),
    (OWL.disjointUnionOf, RDFS.domain, OWL.Class),
    (OWL.disjointWith, RDFS.domain, OWL.Class),
    (OWL.disjointWith, RDFS.range, OWL.Class),
    (OWL.equivalentClass, RDFS.domain, OWL.Class),
    (OWL.equivalentClass, RDFS.range, OWL.Class),
    (OWL.hasKey, RDFS.domain, OWL.Class),
    (OWL.intersectionOf, RDFS.domain, OWL.Class),
    (OWL.onClass, RDFS.range, OWL.Class),
    #    (OWL.onDataRange, RDFS.range, OWL.DataRange),
    (OWL.onDatatype, RDFS.domain, OWL.DataRange),
    (OWL.onDatatype, RDFS.range, OWL.DataRange),
    (OWL.oneOf, RDFS.domain, OWL.Class),
    (OWL.someValuesFrom, RDFS.range, OWL.Class),
    (OWL.unionOf, RDFS.range, OWL.Class),
    #    (OWL.withRestrictions, RDFS.domain, OWL.DataRange)
]

# OWL RL axiomatic triples: combination of the RDFS triples plus the OWL specific ones
OWLRL_Axiomatic_Triples = (
    _OWL_axiomatic_triples_Classes + _OWL_axiomatic_triples_Properties
)

# Note that this is not used anywhere. But I encoded it once and I did not want to remove it...:-)
_OWL_axiomatic_triples_Facets = [
    # langPattern
    (XSD.length, RDF.type, RDF.Property),
    (XSD.maxExclusive, RDF.type, RDF.Property),
    (XSD.maxInclusive, RDF.type, RDF.Property),
    (XSD.maxLength, RDF.type, RDF.Property),
    (XSD.minExclusive, RDF.type, RDF.Property),
    (XSD.minInclusive, RDF.type, RDF.Property),
    (XSD.minLength, RDF.type, RDF.Property),
    (XSD.pattern, RDF.type, RDF.Property),
    (XSD.length, RDFS.domain, RDFS.Resource),
    (XSD.maxExclusive, RDFS.domain, RDFS.Resource),
    (XSD.maxInclusive, RDFS.domain, RDFS.Resource),
    (XSD.maxLength, RDFS.domain, RDFS.Resource),
    (XSD.minExclusive, RDFS.domain, RDFS.Resource),
    (XSD.minInclusive, RDFS.domain, RDFS.Resource),
    (XSD.minLength, RDFS.domain, RDFS.Resource),
    (XSD.pattern, RDFS.domain, RDFS.Resource),
    (XSD.length, RDFS.domain, RDFS.Resource),
    (XSD.maxExclusive, RDFS.range, RDFS.Literal),
    (XSD.maxInclusive, RDFS.range, RDFS.Literal),
    (XSD.maxLength, RDFS.range, RDFS.Literal),
    (XSD.minExclusive, RDFS.range, RDFS.Literal),
    (XSD.minInclusive, RDFS.range, RDFS.Literal),
    (XSD.minLength, RDFS.range, RDFS.Literal),
    (XSD.pattern, RDFS.range, RDFS.Literal),
]

# OWL D-entailment triples (additionally to the RDFS ones), ie, possible subclassing of various extra datatypes
_OWL_D_Axiomatic_Triples_types = [(RDF.PlainLiteral, RDF.type, RDFS.Datatype)]

#
OWL_D_Axiomatic_Triples_subclasses = [
    (XSD.string, RDFS.subClassOf, RDF.PlainLiteral),
    (XSD.normalizedString, RDFS.subClassOf, RDF.PlainLiteral),
    (XSD.token, RDFS.subClassOf, RDF.PlainLiteral),
    (XSD.Name, RDFS.subClassOf, RDF.PlainLiteral),
    (XSD.NCName, RDFS.subClassOf, RDF.PlainLiteral),
    (XSD.NMTOKEN, RDFS.subClassOf, RDF.PlainLiteral),
]

#
OWLRL_Datatypes_Disjointness = [
    (XSD.anyURI, OWL.disjointWith, XSD.base64Binary),
    (XSD.anyURI, OWL.disjointWith, XSD.boolean),
    (XSD.anyURI, OWL.disjointWith, XSD.dateTime),
    (XSD.anyURI, OWL.disjointWith, XSD.decimal),
    (XSD.anyURI, OWL.disjointWith, XSD.double),
    (XSD.anyURI, OWL.disjointWith, XSD.float),
    (XSD.anyURI, OWL.disjointWith, XSD.hexBinary),
    (XSD.anyURI, OWL.disjointWith, XSD.string),
    (XSD.anyURI, OWL.disjointWith, RDF.PlainLiteral),
    (XSD.anyURI, OWL.disjointWith, RDF.XMLLiteral),
    (XSD.base64Binary, OWL.disjointWith, XSD.boolean),
    (XSD.base64Binary, OWL.disjointWith, XSD.dateTime),
    (XSD.base64Binary, OWL.disjointWith, XSD.decimal),
    (XSD.base64Binary, OWL.disjointWith, XSD.double),
    (XSD.base64Binary, OWL.disjointWith, XSD.float),
    (XSD.base64Binary, OWL.disjointWith, XSD.hexBinary),
    (XSD.base64Binary, OWL.disjointWith, XSD.string),
    (XSD.base64Binary, OWL.disjointWith, RDF.PlainLiteral),
    (XSD.base64Binary, OWL.disjointWith, RDF.XMLLiteral),
    (XSD.boolean, OWL.disjointWith, XSD.dateTime),
    (XSD.boolean, OWL.disjointWith, XSD.decimal),
    (XSD.boolean, OWL.disjointWith, XSD.double),
    (XSD.boolean, OWL.disjointWith, XSD.float),
    (XSD.boolean, OWL.disjointWith, XSD.hexBinary),
    (XSD.boolean, OWL.disjointWith, XSD.string),
    (XSD.boolean, OWL.disjointWith, RDF.PlainLiteral),
    (XSD.boolean, OWL.disjointWith, RDF.XMLLiteral),
    (XSD.dateTime, OWL.disjointWith, XSD.decimal),
    (XSD.dateTime, OWL.disjointWith, XSD.double),
    (XSD.dateTime, OWL.disjointWith, XSD.float),
    (XSD.dateTime, OWL.disjointWith, XSD.hexBinary),
    (XSD.dateTime, OWL.disjointWith, XSD.string),
    (XSD.dateTime, OWL.disjointWith, RDF.PlainLiteral),
    (XSD.dateTime, OWL.disjointWith, RDF.XMLLiteral),
    (XSD.decimal, OWL.disjointWith, XSD.double),
    (XSD.decimal, OWL.disjointWith, XSD.float),
    (XSD.decimal, OWL.disjointWith, XSD.hexBinary),
    (XSD.decimal, OWL.disjointWith, XSD.string),
    (XSD.decimal, OWL.disjointWith, RDF.PlainLiteral),
    (XSD.decimal, OWL.disjointWith, RDF.XMLLiteral),
    (XSD.double, OWL.disjointWith, XSD.float),
    (XSD.double, OWL.disjointWith, XSD.hexBinary),
    (XSD.double, OWL.disjointWith, XSD.string),
    (XSD.double, OWL.disjointWith, RDF.PlainLiteral),
    (XSD.double, OWL.disjointWith, RDF.XMLLiteral),
    (XSD.float, OWL.disjointWith, XSD.hexBinary),
    (XSD.float, OWL.disjointWith, XSD.string),
    (XSD.float, OWL.disjointWith, RDF.PlainLiteral),
    (XSD.float, OWL.disjointWith, RDF.XMLLiteral),
    (XSD.hexBinary, OWL.disjointWith, XSD.string),
    (XSD.hexBinary, OWL.disjointWith, RDF.PlainLiteral),
    (XSD.hexBinary, OWL.disjointWith, RDF.XMLLiteral),
    (XSD.string, OWL.disjointWith, RDF.XMLLiteral),
]

# OWL RL D Axiomatic triples: combination of the RDFS ones, plus some extra statements on ranges and domains, plus
# some OWL specific datatypes
OWLRL_D_Axiomatic_Triples = (
    RDFS_D_Axiomatic_Triples
    + _OWL_D_Axiomatic_Triples_types
    + OWL_D_Axiomatic_Triples_subclasses
    + OWLRL_Datatypes_Disjointness
)



# -*- coding: utf-8 -*-
#
"""
The combined closure: performing *both* the OWL 2 RL and RDFS closures.

The two are very close but there are some rules in RDFS that are not in OWL 2 RL (eg, the axiomatic
triples concerning the container membership properties). Using this closure class the
OWL 2 RL implementation becomes a full extension of RDFS.

**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""

__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

from typing import Union

from rdflib import Graph
from rdflib.namespace import OWL, RDF, RDFS

from owlrl.RDFSClosure import RDFS_Semantics
from owlrl.OWLRL import OWLRL_Semantics

######################################################################################################


# noinspection PyPep8Naming
class RDFS_OWLRL_Semantics(RDFS_Semantics, OWLRL_Semantics):
    """
    Common subclass of the RDFS and OWL 2 RL semantic classes. All methods simply call back
    to the functions in the superclasses. This may lead to some unnecessary duplication of terms
    and rules, but it is not so bad. Also, the additional identification defined for OWL Full,
    ie, Resource being the same as Thing and OWL and RDFS classes being identical are added to the
    triple store.

    Note that this class is also a possible user extension point: subclasses can be created that
    extend the standard functionality by extending this class. This class *always*} performs RDFS inferences.
    Subclasses have to set the :code:`self.rdfs` flag explicitly to the requested value if that is to be controlled.

    :param graph: The RDF graph to be extended.
    :type graph: :class:`rdflib.Graph`

    :param axioms: Whether (non-datatype) axiomatic triples should be added or not.
    :type axioms: bool

    :param daxioms: Whether datatype axiomatic triples should be added or not.
    :type daxioms: bool

    :param rdfs: Placeholder flag (used in subclassed only, it is always defaulted to True in this class)
    :type rdfs: bool

    :var full_binding_triples: Additional axiom type triples that are added to the combined semantics; these 'bind'
        the RDFS and the OWL worlds together.

    :var rdfs: (bool) Whether RDFS inference is to be performed or not. In this class instance the value is *always*
        :code:`True`, subclasses may explicitly change it at initialization time.
    :type rdfs: bool
    """

    full_binding_triples = [
        (OWL.Thing, OWL.equivalentClass, RDFS.Resource),
        (RDFS.Class, OWL.equivalentClass, OWL.Class),
        (OWL.DataRange, OWL.equivalentClass, RDFS.Datatype),
    ]

    def __init__(self, graph: Graph, axioms, daxioms, rdfs: bool = True, destination: Union[None, Graph] = None):
        """
        @param graph: the RDF graph to be extended
        @type graph: rdflib.Graph
        @param axioms: whether (non-datatype) axiomatic triples should be added or not
        @type axioms: bool
        @param daxioms: whether datatype axiomatic triples should be added or not
        @type daxioms: bool
        @param rdfs: placeholder flag (used in subclassed only, it is always defaulted to True in this class)
        @type rdfs: boolean
        @param destination: the destination graph to which the results are written. If None, use the source graph.
        @type destination: rdflib.Graph
        """
        OWLRL_Semantics.__init__(self, graph, axioms, daxioms, rdfs=rdfs, destination=destination)
        RDFS_Semantics.__init__(self, graph, axioms, daxioms, rdfs=rdfs, destination=destination)
        self.rdfs = True

    # noinspection PyMethodMayBeStatic
    @staticmethod
    def add_new_datatype(
        uri,
        conversion_function,
        datatype_list,
        subsumption_dict=None,
        subsumption_key=None,
        subsumption_list=None,
    ):
        """
        If an extension wants to add new datatypes, this method should be invoked at initialization time.

        :param uri: URI for the new datatypes, like owl_ns["Rational"].

        :param conversion_function: A function converting the lexical representation of the datatype to a Python value,
            possibly raising an exception in case of unsuitable lexical form.

        :param datatype_list: List of datatypes already in use that has to be checked.
        :type datatype_list: list

        :param subsumption_dict: Dictionary of subsumption hierarchies (indexed by the datatype URI-s).
        :type subsumption_dict: dict

        :param subsumption_key: Key in the dictionary, if None, the uri parameter is used.
        :type subsumption_key: str

        :param subsumption_list: List of subsumptions associated to a subsumption key (ie, all datatypes that are
            superclasses of the new datatype).
        :type subsumption_list: list
        """
        from .DatatypeHandling import AltXSDToPYTHON, use_Alt_lexical_conversions

        if datatype_list:
            datatype_list.append(uri)

        if subsumption_dict and subsumption_list:
            if subsumption_key:
                subsumption_dict[subsumption_key] = subsumption_list
            else:
                subsumption_dict[uri] = subsumption_list

        AltXSDToPYTHON[uri] = conversion_function
        use_Alt_lexical_conversions()

    def post_process(self):
        """
        Do some post-processing step. This method when all processing is done, but before handling possible
        errors (I.e., the method can add its own error messages). By default, this method is empty, subclasses
        can add content to it by overriding it.
        """
        OWLRL_Semantics.post_process(self)

    def rules(self, t, cycle_num):
        """
        :param t: A triple (in the form of a tuple).
        :type t: tuple

        :param cycle_num: Which cycle are we in, starting with 1. This value is forwarded to all local rules; it is
            also used locally to collect the bnodes in the graph.
        :type cycle_num: int
        """
        OWLRL_Semantics.rules(self, t, cycle_num)
        if self.rdfs:
            RDFS_Semantics.rules(self, t, cycle_num)

    def add_axioms(self):
        if self.rdfs:
            RDFS_Semantics.add_axioms(self)
        OWLRL_Semantics.add_axioms(self)

    def add_d_axioms(self):
        if self.rdfs:
            RDFS_Semantics.add_d_axioms(self)
        OWLRL_Semantics.add_d_axioms(self)

    def one_time_rules(self):
        """Adds some extra axioms and calls for the d_axiom part of the OWL Semantics."""
        for t in self.full_binding_triples:
            self.store_triple(t)

        # Note that the RL one time rules include the management of datatype which is a true superset
        # of the rules in RDFS. It is therefore unnecessary to add those even self.rdfs is True.
        OWLRL_Semantics.one_time_rules(self)



# -*- coding: utf-8 -*-
#
"""
Lists of XSD datatypes and their mutual relationships

**Requires**: `RDFLib`_, 4.0.0 and higher.

.. _RDFLib: https://github.com/RDFLib/rdflib

**License**: This software is available for use under the `W3C Software License`_.

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""
__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

from rdflib.namespace import RDF, RDFS, XSD

# The basic XSD types used everywhere; this means not the complete set of day/month types
_Common_XSD_Datatypes = [
    XSD.integer,
    XSD.decimal,
    XSD.nonNegativeInteger,
    XSD.nonPositiveInteger,
    XSD.negativeInteger,
    XSD.positiveInteger,
    XSD.long,
    XSD.int,
    XSD.short,
    XSD.byte,
    XSD.unsignedLong,
    XSD.unsignedInt,
    XSD.unsignedShort,
    XSD.unsignedByte,
    XSD.float,
    XSD.double,
    XSD.string,
    XSD.normalizedString,
    XSD.token,
    XSD.language,
    XSD.Name,
    XSD.NCName,
    XSD.NMTOKEN,
    XSD.boolean,
    XSD.hexBinary,
    XSD.base64Binary,
    XSD.anyURI,
    XSD.dateTimeStamp,
    XSD.dateTime,
    XSD.time,
    XSD.date,
    RDFS.Literal,
    RDF.XMLLiteral,
    RDF.HTML,
    RDF.langString,
]

# RDFS Datatypes: the basic ones plus the complete set of day/month ones
RDFS_Datatypes = _Common_XSD_Datatypes + [
    XSD.gYearMonth,
    XSD.gMonthDay,
    XSD.gYear,
    XSD.gDay,
    XSD.gMonth,
]

# OWL RL Datatypes: the basic ones plus plain literal
OWL_RL_Datatypes = _Common_XSD_Datatypes + [RDF.PlainLiteral]

# XSD Datatype subsumptions
_Common_Datatype_Subsumptions = {
    XSD.dateTimeStamp: [XSD.dateTime],
    XSD.integer: [XSD.decimal],
    XSD.long: [XSD.integer, XSD.decimal],
    XSD.int: [XSD.long, XSD.integer, XSD.decimal],
    XSD.short: [
        XSD.int,
        XSD.long,
        XSD.integer,
        XSD.decimal,
    ],
    XSD.byte: [
        XSD.short,
        XSD.int,
        XSD.long,
        XSD.integer,
        XSD.decimal,
    ],
    XSD.nonNegativeInteger: [XSD.integer, XSD.decimal],
    XSD.positiveInteger: [
        XSD.nonNegativeInteger,
        XSD.integer,
        XSD.decimal,
    ],
    XSD.unsignedLong: [
        XSD.nonNegativeInteger,
        XSD.integer,
        XSD.decimal,
    ],
    XSD.unsignedInt: [
        XSD.unsignedLong,
        XSD.nonNegativeInteger,
        XSD.integer,
        XSD.decimal,
    ],
    XSD.unsignedShort: [
        XSD.unsignedInt,
        XSD.unsignedLong,
        XSD.nonNegativeInteger,
        XSD.integer,
        XSD.decimal,
    ],
    XSD.unsignedByte: [
        XSD.unsignedShort,
        XSD.unsignedInt,
        XSD.unsignedLong,
        XSD.nonNegativeInteger,
        XSD.integer,
        XSD.decimal,
    ],
    XSD.nonPositiveInteger: [XSD.integer, XSD.decimal],
    XSD.negativeInteger: [
        XSD.nonPositiveInteger,
        XSD.integer,
        XSD.decimal,
    ],
    XSD.normalizedString: [XSD.string],
    XSD.token: [XSD.normalizedString, XSD.string],
    XSD.language: [XSD.token, XSD.normalizedString, XSD.string],
    XSD.Name: [XSD.token, XSD.normalizedString, XSD.string],
    XSD.NCName: [
        XSD.Name,
        XSD.token,
        XSD.normalizedString,
        XSD.string,
    ],
    XSD.NMTOKEN: [
        XSD.Name,
        XSD.token,
        XSD.normalizedString,
        XSD.string,
    ],
}

# RDFS Datatype subsumptions: at the moment, there is no extra to XSD
RDFS_Datatype_Subsumptions = _Common_Datatype_Subsumptions

# OWL Datatype subsumptions: at the moment, there is no extra to XSD
OWL_Datatype_Subsumptions = _Common_Datatype_Subsumptions



# -*- coding: utf-8 -*-
#
"""
This module is a brute force implementation of the 'finite' version of `RDFS semantics`_ and of `OWL 2 RL`_ on the top
of RDFLib (with some caveats, see below). Some extensions to these are also implemented.

.. _RDFS semantics: http://www.w3.org/TR/rdf-mt/
.. _OWL 2 RL: http://www.w3.org/TR/owl2-profiles/#Reasoning_in_OWL_2_RL_and_RDF_Graphs_using_Rules

Brute force means that, in all cases, simple forward chaining rules are used to extend (recursively) the incoming graph
with all triples that the rule sets permit (ie, the "deductive closure" of the graph is computed).
There is an extra options whether the axiomatic triples are added to the graph (prior to the forward chaining step).
These, typically set the domain and range for properties or define some core classes.
In the case of RDFS, the implementation uses a 'finite' version of the axiomatic triples only (as proposed, for example,
by Herman ter Horst). This means that it adds only those :code:`rdf:_i` type predicates that do appear in the original graph,
thereby keeping this step finite. For OWL 2 RL, OWL 2 does not define axiomatic triples formally; but they can be
deduced from the `OWL 2 RDF Based Semantics`_ document and are listed in Appendix 6 (though informally).

.. _OWL 2 RDF Based Semantics: http://www.w3.org/TR/owl2-rdf-based-semantics/

.. note:: This implementation adds only those triples that refer to OWL terms that are meaningful for the OWL 2 RL case.


Package Entry Points
====================

The main entry point to the package is via the :class:`.DeductiveClosure` class. This class should be
initialized to control the parameters of the deductive closure; the forward chaining is done via the
L{expand<owlrl.DeductiveClosure.expand>} method.
The simplest way to use the package from an RDFLib application is as follows::

    graph = Graph()                                 # creation of an RDFLib graph
    ...
    ...                                             # normal RDFLib application, eg, parsing RDF data
    ...
    DeductiveClosure(OWLRL_Semantics).expand(graph) # calculate an OWL 2 RL deductive closure of graph
                                                    # without axiomatic triples

The first argument of the :class:`.DeductiveClosure` initialization can be replaced by other classes, providing different
types of deductive closure; other arguments are also possible. For example::

 DeductiveClosure(OWLRL_Extension, rdfs_closure = True, axiomatic_triples = True, datatype_axioms = True).expand(graph)

This will calculate the deductive closure including RDFS and some extensions to OWL 2 RL, and with all possible axiomatic
triples added to the graph (this is about the maximum the package can do…)

The same instance of :class:`.DeductiveClosure` can be used for several graph expansions. In other words, the
expand function does not change any state.

For convenience, a second entry point to the package is provided in the form of a function called
:func:`owlrl.convert_graph`, that expects a directory with various options, including a file name. The function
parses the file, creates the expanded graph, and serializes the result into RDF/XML or Turtle. This function is
particularly useful as an entry point for a CGI call (where the HTML form parameters are in a directory) and is easy to
use with a command line interface. The package distribution contains an example for both.

There are major closure type (ie, semantic closure possibilities); these can be controlled through the appropriate
parameters of the :class:`.DeductiveClosure` class:

    * using the :class:`.RDFS_Semantics` class, implementing the `RDFS semantics`_.

    .. _RDFS semantics: http://www.w3.org/TR/rdf-mt/

    * using the :class:`.OWLRL.OWLRL_Semantics` class, implementing the `OWL 2 RL`_.

    .. _OWL 2 RL: http://www.w3.org/TR/owl2-profiles/#Reasoning_in_OWL_2_RL_and_RDF_Graphs_using_Rules

    * using :class:`.CombinedClosure.RDFS_OWLRL_Semantics` class, implementing a combined semantics of `RDFS semantics`_ and `OWL 2 RL`_.

    .. _RDFS semantics: http://www.w3.org/TR/rdf-mt/
    .. _OWL 2 RL: http://www.w3.org/TR/owl2-profiles/#Reasoning_in_OWL_2_RL_and_RDF_Graphs_using_Rules

In all three cases there are other dimensions that can control the exact closure being generated:

    * for convenience, the so called axiomatic triples (see, eg, the `axiomatic triples in RDFS`_ are, by default, I{not} added to the graph closure to reduce the number of generated triples. These can be controlled through a separate initialization argument.

    .. _axiomatic triples in RDFS: http://www.w3.org/TR/rdf-mt/#rdfs_interp

    * similarly, the axiomatic triples for D-entailment are separated.

Some Technical/implementation aspects
=====================================

The core processing is done in the in the :class:`.Closure.Core` class, which is subclassed by the
:class:`.RDFSClosure.RDFS_Semantics` and the :class:`.OWLRL.OWLRL_Semantics` classes (these two are then, on their turn,
subclassed by the :class:`.CombinedClosure.RDFS_OWLRL_Semantics` class). The core implements the core functionality of
cycling through the rules, whereas the rules themselves are defined and implemented in the subclasses. There are also
methods that are executed only once either at the beginning or at the end of the full processing cycle. Adding axiomatic
triples is handled separately, which allows a finer user control over these features.

The OWL specification includes references to datatypes that are not in the core RDFS specification, consequently not
directly implemented by RDFLib. These are added in a separate module of the package.


Problems with Literals with datatypes
-------------------------------------

The current distribution of RDFLib is fairly poor in handling datatypes, particularly in checking whether a lexical form
of a literal is "proper" as for its declared datatype. A typical example is ::

  "-1234"^^xsd:nonNegativeInteger

which should not be accepted as valid literal. Because the requirements of OWL 2 RL are much stricter in this respect,
an alternative set of datatype handling (essentially, conversions) had to be implemented (see the :py:mod:`.XsdDatatypes`
module).

The :class:`.DeductiveClosure` class has an additional instance variable whether
the default RDFLib conversion routines should be exchanged against the new ones. If this flag is set to True and
instance creation (this is the default), then the conversion routines are set back to the originals once the expansion
is complete, thereby avoiding to influence older application that may not work properly with the new set of conversion
routines.

If the user wants to use these alternative lexical conversions everywhere in the application, then
the :py:meth:`.DeductiveClosure.use_improved_datatypes_conversions` method can be invoked.
That method changes the conversion routines and, from that point on, all usage of :class:`.DeductiveClosure` instances
will use the improved conversion methods without resetting them. Ie, the code structure can be something like::

    DeductiveClosure().use_improved_datatypes_conversions()
    ... RDFLib application
    DeductiveClosure().expand(graph)
    ...


The default situation can be set back using the
:py:meth:`.DeductiveClosure.use_rdflib_datatypes_conversions` call.

It is, however, not *required* to use these methods at all. I.e., the user can use::

    DeductiveClosure(improved_datatypes=False).expand(graph)


which will result in a proper graph expansion except for the datatype specific comparisons which will be incomplete.


**Requires**:
    * `RDFLib`_, 4.0.0 and higher.

    .. _RDFLib: https://github.com/RDFLib/rdflib

    * `rdflib_jsonld`_

    .. _rdflib_jsonld: https://github.com/RDFLib/rdflib-jsonld

**License**: This software is available for use under the `W3C Software License`_

.. _W3C Software License: http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231

**Organization**: `World Wide Web Consortium`_

.. _World Wide Web Consortium: http://www.w3.org

**Author**: `Ivan Herman`_

.. _Ivan Herman: http://www.w3.org/People/Ivan/

"""

# Examples: LangString is disjoint from String
__version__ = "7.1.2"
__author__ = "Ivan Herman"
__contact__ = "Ivan Herman, ivan@w3.org"
__license__ = "W3C® SOFTWARE NOTICE AND LICENSE, http://www.w3.org/Consortium/Legal/2002/copyright-software-20021231"

from typing import Union

# noinspection PyPackageRequirements,PyPackageRequirements,PyPackageRequirements
import rdflib
from rdflib import Graph, Literal

from . import DatatypeHandling, Closure
from .OWLRLExtras import OWLRL_Extension, OWLRL_Extension_Trimming
from .OWLRL import OWLRL_Semantics
from .RDFSClosure import RDFS_Semantics
from .CombinedClosure import RDFS_OWLRL_Semantics
from rdflib.namespace import OWL

RDFXML = "xml"
TURTLE = "turtle"
JSON = "json"
AUTO = "auto"
RDFA = "rdfa"


# noinspection PyShadowingBuiltins
def __parse_input(iformat, inp, graph):
    """Parse the input into the graph, possibly checking the suffix for the format.

    @param iformat: input format; can be one of L{AUTO}, L{TURTLE}, or L{RDFXML}. L{AUTO} means that the suffix of the
    file name or URI will decide: '.ttl' means Turtle, RDF/XML otherwise.
    @param inp: input file; anything that RDFLib accepts in that position (URI, file name, file object). If '-',
    standard input is used.
    @param graph: the RDFLib Graph instance to parse into.
    """
    if iformat == AUTO:
        if inp == "-":
            format = "turtle"
        else:
            if inp.endswith(".ttl") or inp.endswith(".n3"):
                format = "turtle"
            elif inp.endswith(".json") or inp.endswith(".jsonld"):
                format = "json-ld"
            elif inp.endswith(".html"):
                format = "rdfa1.1"
            else:
                format = "xml"
    elif iformat == TURTLE:
        format = "turtle"
    elif iformat == RDFA:
        format = "rdfa1.1"
    elif iformat == RDFXML:
        format = "xml"
    elif iformat == JSON:
        format = "json-ld"
    else:
        raise Exception("Unknown input syntax")

    if inp == "-":
        # standard input is used
        import sys

        source = sys.stdin
    else:
        source = inp
    graph.parse(source, format=format)


def interpret_owl_imports(iformat, graph):
    """
    Interpret the owl import statements. Essentially, recursively merge with all the objects in the owl import
    statement, and remove the corresponding triples from the graph.

    This method can be used by an application prior to expansion. It is *not* done by the :class:`.DeductiveClosure`
    class.

    :param iformat: Input format; can be one of :code:`AUTO`, :code:`TURTLE`, or :code:`RDFXML`. :code:`AUTO` means that the suffix of the file name or URI will decide: '.ttl' means Turtle, RDF/XML otherwise.
    :type iformat: str
    :param graph: The RDFLib Graph instance to parse into.
    :type graph: :class:`rdflib.graph.Graph`
    """
    while True:
        # 1. collect the import statements:
        all_imports = [t for t in graph.triples((None, OWL.imports, None))]
        if len(all_imports) == 0:
            # no import statement whatsoever, we can go on...
            return
        # 2. remove all the import statements from the graph
        for t in all_imports:
            graph.remove(t)
        # 3. get all the imported vocabularies and import them
        for (s, p, uri) in all_imports:
            # this is not 100% kosher. The expected object for an import statement is a URI. However,
            # on local usage, a string would also make sense, so I do that one, too
            if isinstance(uri, Literal):
                __parse_input(iformat, str(uri), graph)
            else:
                __parse_input(iformat, uri, graph)
            # 4. start all over again to see if import statements have been imported


def return_closure_class(owl_closure, rdfs_closure, owl_extras, trimming=False):
    """
    Return the right semantic extension class based on three possible choices (this method is here to help potential
    users, the result can be fed into a :class:`DeductiveClosure` instance at initialization).

    :param owl_closure: Whether OWL 2 RL deductive closure should be calculated.
    :type owl_closure: bool

    :param rdfs_closure: Whether RDFS deductive closure should be calculated. In case :code:`owl_closure==True`, this
        parameter should also be used in the initialization of :class:`DeductiveClosure`.
    :type rdfs_closure: bool

    :param owl_extras: Whether the extra possibilities (rational datatype, etc) should be added to an OWL 2 RL
        deductive closure. This parameter has no effect in case :code:`owl_closure==False`.
    :type owl_extras: bool

    :param trimming: Whether extra trimming is done on the OWL RL + Extension output.
    :type trimming: bool

    :return: Deductive class reference or None.
    :rtype: :class:`.DeductiveClosure` or None
    """
    if owl_closure:
        if owl_extras:
            if trimming:
                return OWLRL_Extension_Trimming
            else:
                return OWLRL_Extension
        else:
            if rdfs_closure:
                return RDFS_OWLRL_Semantics
            else:
                return OWLRL_Semantics
    elif rdfs_closure:
        return RDFS_Semantics
    else:
        return None


# noinspection PyCallingNonCallable
class DeductiveClosure:
    """
    Entry point to generate the deductive closure of a graph. The exact choice deductive
    closure is controlled by a class reference. The important initialization parameter is the :code:`closure_class`, a Class
    object referring to a subclass of :class:`.Closure.Core`. Although this package includes a number of such subclasses
    :class:`.OWLRL_Semantics`, :class:`.RDFS_Semantics`, :class:`.RDFS_OWLRL_Semantics`, and :class:`.OWLRL_Extension`, the user can use his/her
    own if additional rules are implemented.

    Note that :code:`owl:imports` statements are *not* interpreted in this class, that has to be done beforehand on the graph
    that is to be expanded.

    :param closure_class: A closure class reference.
    :type closure_class: subclass of :class:`.Closure.Core`

    :param improved_datatypes: Whether the improved set of lexical-to-Python conversions should be used for datatype handling. See the introduction for more details. Default: True.
    :type improved_datatypes: bool

    :param rdfs_closure: Whether the RDFS closure should also be executed. Default: False.
    :type rdfs_closure: bool

    :param axiomatic_triples: Whether relevant axiomatic triples are added before chaining, except for datatype axiomatic triples. Default: False.
    :type axiomatic_triples: bool

    :param datatype_axioms: Whether further datatype axiomatic triples are added to the output. Default: false.
    :type datatype_axioms: bool

    :var improved_datatype_generic: Whether the improved set of lexical-to-Python conversions should be used for datatype handling *in general*, I.e., not only for a particular instance and not only for inference purposes. Default: False.
    :type improved_datatype_generic: bool
    """

    # This is the original set of param definitions in the class definition
    #
    # @ivar rdfs_closure: Whether the RDFS closure should also be executed. Default: False.
    # @type rdfs_closure: boolean
    # @ivar axiomatic_triples: Whether relevant axiomatic triples are added before chaining, except for datatype axiomatic
    # triples. Default: False.
    # @type axiomatic_triples: boolean
    # @ivar datatype_axioms: Whether further datatype axiomatic triples are added to the output. Default: false.
    # @type datatype_axioms: boolean
    # @ivar closure_class: the class instance used to expand the graph
    # @type closure_class: L{Closure.Core}
    # @cvar improved_datatype_generic: Whether the improved set of lexical-to-Python conversions should be used for
    # datatype handling I{in general}, ie, not only for a particular instance and not only for inference purposes.
    # Default: False.
    # @type improved_datatype_generic: boolean

    improved_datatype_generic = False

    def __init__(
        self,
        closure_class,
        improved_datatypes=True,
        rdfs_closure=False,
        axiomatic_triples=False,
        datatype_axioms=False,
    ):
        # This is the original set of param definitions in the __init__
        #
        # @param closure_class: a closure class reference.
        # @type closure_class: subclass of L{Closure.Core}
        # @param rdfs_closure: whether RDFS rules are executed or not
        # @type rdfs_closure: boolean
        # @param axiomatic_triples: Whether relevant axiomatic triples are added before chaining, except for datatype
        # axiomatic triples. Default: False.
        # @type axiomatic_triples: boolean
        # @param datatype_axioms: Whether further datatype axiomatic triples are added to the output. Default: false.
        # @type datatype_axioms: boolean
        # @param improved_datatypes: Whether the improved set of lexical-to-Python conversions should be used for
        # datatype handling. See the introduction for more details. Default: True.
        # @type improved_datatypes: boolean

        if closure_class is None:
            self.closure_class = None
        else:
            if not isinstance(closure_class, type):
                raise ValueError("The closure type argument must be a class reference")
            else:


                self.closure_class = closure_class
        self.axiomatic_triples = axiomatic_triples
        self.datatype_axioms = datatype_axioms
        self.rdfs_closure = rdfs_closure
        self.improved_datatypes = improved_datatypes

    def expand(self, graph: Graph, destination: Union[None, Graph] = None):
        """
        Expand the graph using forward chaining, and with the relevant closure type.

        :param graph: The RDF graph.
        :type graph: :class:`rdflib.Graph`
        :param destination: The RDF graph to which the results are written. If not specified, the graph is modified in-place.
        :type destination: :class:`rdflib.Graph`
        """
        if (not DeductiveClosure.improved_datatype_generic) and self.improved_datatypes:
            DatatypeHandling.use_Alt_lexical_conversions()

        if self.closure_class is not None:
            self.closure_class(
                graph,
                self.axiomatic_triples,
                self.datatype_axioms,
                rdfs=self.rdfs_closure,
                destination=destination
            ).closure()

        if (not DeductiveClosure.improved_datatype_generic) and self.improved_datatypes:
            DatatypeHandling.use_RDFLib_lexical_conversions()

    @staticmethod
    def use_improved_datatypes_conversions():
        """
        Switch the system to use the improved datatype conversion routines.
        """
        DeductiveClosure.improved_datatype_generic = True
        DatatypeHandling.use_Alt_lexical_conversions()

    @staticmethod
    def use_rdflib_datatypes_conversions():
        """
        Switch the system to use the generic (RDFLib) datatype conversion routines
        """
        DeductiveClosure.improved_datatype_generic = False
        DatatypeHandling.use_RDFLib_lexical_conversions()


###############################################################################################################


# noinspection PyPep8Naming,PyBroadException,PyBroadException,PyBroadException
def convert_graph(options, closureClass=None):
    """
    Entry point for external scripts (CGI or command line) to parse an RDF file(s), possibly execute OWL and/or RDFS
    closures, and serialize back the result in some format.

    Note that this entry point can be used requiring no entailment at all;
    because both the input and the output format for the package can be RDF/XML or Turtle, such usage would
    simply mean a format conversion.

    If OWL 2 RL processing is required, that also means that the :code:`owl:imports` statements are interpreted. I.e.,
    ontologies can be spread over several files. Note, however, that the output of the process would then include all
    imported ontologies, too.

    :param options: Object with specific attributes.
    :type options: object

    :param options.sources: List of uris or file names for the source data; for each one if the name ends with 'ttl', it
        is considered to be turtle, RDF/XML otherwise (this can be overwritten by the options.iformat, though)
    :type options.sources: list

    :param options.text: Direct Turtle encoding of a graph as a text string (useful, eg, for a CGI call using a text
        field).
    :type options.text: str

    :param options.owlClosure: Can be yes or no.
    :type options.owlClosure: bool

    :param options.rdfsClosure: Can be yes or no.
    :type options.rdfsClosure: bool

    :param options.owlExtras: Can be yes or no; whether the extra rules beyond OWL 2 RL are used or not.
    :type options.owlExtras: bool

    :param options.axioms: Whether relevant axiomatic triples are added before chaining (can be a boolean, or the
        strings "yes" or "no").
    :type options.axioms: bool

    :param options.daxioms: Further datatype axiomatic triples are added to the output (can be a boolean, or the strings
        "yes" or "no").
    :type options.daxioms: bool

    :param options.format: Output format, can be "turtle" or "rdfxml".
    :type options.format: str

    :param options.iformat: Input format, can be "turtle", "rdfa", "json", "rdfxml", or "auto". "auto" means that the
        suffix of the file is considered: '.ttl'. '.html', 'json' or '.jsonld' respectively with 'xml' as a fallback.
    :type options.iformat: str

    :param options.trimming: Whether the extension to OWLRL should also include trimming.
    :type options.trimming: bool

    :param closureClass: Explicit class reference. If set, this overrides the various different other options to be
        used as an extension.
    :type closureClass: TODO(edmond.chuc@csiro.au): What class is this supposed to be?
    """

    # Original parameter definitions from old documentation.
    #
    # @param options: object with specific attributes, namely:
    #   - options.sources: list of uris or file names for the source data; for each one if the name ends with 'ttl', it is
    #     considered to be turtle, RDF/XML otherwise (this can be overwritten by the options.iformat, though)
    #   - options.text: direct Turtle encoding of a graph as a text string (useful, eg, for a CGI call using a text field)
    #   - options.owlClosure: can be yes or no
    #   - options.rdfsClosure: can be yes or no
    #   - options.owlExtras: can be yes or no; whether the extra rules beyond OWL 2 RL are used or not.
    #   - options.axioms: whether relevant axiomatic triples are added before chaining (can be a boolean, or the strings
    #     "yes" or "no")
    #   - options.daxioms: further datatype axiomatic triples are added to the output (can be a boolean, or the strings
    #     "yes" or "no")
    #   - options.format: output format, can be "turtle" or "rdfxml"
    #   - options.iformat: input format, can be "turtle", "rdfa", "json", "rdfxml", or "auto". "auto" means that the
    #     suffix of the file is considered: '.ttl'. '.html', 'json' or '.jsonld' respectively with 'xml' as a fallback
    #   - options.trimming: whether the extension to OWLRL should also include trimming
    # @param closureClass: explicit class reference. If set, this overrides the various different other options to be
    #     used as an extension.

    def __check_yes_or_true(opt):
        return (
            opt is True
            or opt == "yes"
            or opt == "Yes"
            or opt == "True"
            or opt == "true"
        )

    import warnings

    warnings.filterwarnings("ignore")
    if len(options.sources) == 0 and (
        options.text is None or len(options.text.strip()) == 0
    ):
        raise Exception("No graph specified either via a URI or text")

    graph = Graph()

    # Just to be sure that this attribute does not create issues with older versions of the service...
    # the try statement should be removed, eventually...
    iformat = "auto"
    try:
        iformat = options.iformat
    except:
        # exception can be raised if that attribute is not used at all, true for older versions
        pass

    # similar measure with the possible usage of the 'source' options
    try:
        if options.source is not None:
            options.sources.append(options.source)
    except:
        # exception can be raised if that attribute is not used at all, true for newer versions
        pass

    # Get the sources first. Note that a possible error is filtered out, namely to process the same file twice. This is
    # done by turning the input arguments into a set...
    for inp in set(options.sources):
        __parse_input(iformat, inp, graph)

    # add the possible extra text (ie, the text input on the HTML page)
    if options.text is not None:
        graph.parse(data=options.text, format="n3")

    # Get all the options right
    # noinspection PyPep8Naming
    owlClosure = __check_yes_or_true(options.owlClosure)
    # noinspection PyPep8Naming
    rdfsClosure = __check_yes_or_true(options.rdfsClosure)
    # noinspection PyPep8Naming
    owlExtras = __check_yes_or_true(options.owlExtras)
    try:
        trimming = __check_yes_or_true(options.trimming)
    except:
        trimming = False
    axioms = __check_yes_or_true(options.axioms)
    daxioms = __check_yes_or_true(options.daxioms)

    if owlClosure:
        interpret_owl_imports(iformat, graph)

    # @@@@ some smarter choice should be used later to decide what the closure class is!!! That should
    # also control the import management. Eg, if the superclass includes OWL...
    if closureClass is not None:
        closure_class = closureClass
    else:
        closure_class = return_closure_class(
            owlClosure, rdfsClosure, owlExtras, trimming
        )

    DeductiveClosure(
        closure_class,
        improved_datatypes=True,
        rdfs_closure=rdfsClosure,
        axiomatic_triples=axioms,
        datatype_axioms=daxioms,
    ).expand(graph)

    if options.format == "rdfxml":
        return graph.serialize(format="pretty-xml")
    elif options.format == "json":
        return graph.serialize(format="json-ld")
    else:
        return graph.serialize(format="turtle")
